type stp = Base of string | Prop | Ar of (stp * stp)

type trm =
    Name of string * stp
  | False | Imp | Forall of stp | Eq of stp | Choice of stp
  | True | Neg | Or | And | Iff | Exists of stp (*** These are normalized away. ***)
  | DB of int * stp
  | Lam of stp * trm
  | Ap of trm * trm

let rec varint_binout c i =
  if i < 0 then
    begin
      output_byte c 255;
      varint_binout c (- i)
    end
  else if i <= 252 then
    output_byte c i
  else if i < 65536 then
    begin
      output_byte c 253;
      output_byte c (i land 255);
      output_byte c ((i lsr 8) land 255)
    end
  else
    begin
      output_byte c 254;
      output_byte c (i land 255);
      output_byte c ((i lsr 8) land 255);
      output_byte c ((i lsr 16) land 255);
      output_byte c ((i lsr 24) land 255)
    end

      (*** strings > 255 chars are truncated to the first 255 chars ***)
let string_binout c x =
  let n = String.length x in
  if n < 256 then
    begin
      output_byte c n;
      for i = 0 to n-1 do
	output_byte c (Char.code (x.[i]))
      done;
    end
  else
    begin
      output_byte c 255;
      for i = 0 to 254 do
	output_byte c (Char.code (x.[i]))
      done
    end

let rec stp_binout c a =
  match a with
  | Prop ->
      output_byte c 1
  | Base(x) ->
      output_byte c 2;
      string_binout c x
  | Ar(a1,a2) ->
      output_byte c 3;
      stp_binout c a1;
      stp_binout c a2

let rec trm_binout c m =
  match m with
  | Name(x,a) ->
      output_byte c 1;
      string_binout c x;
      stp_binout c a
  | False -> output_byte c 2
  | Imp -> output_byte c 3
  | Forall(a) ->
      output_byte c 4;
      stp_binout c a
  | Eq(a) ->
      output_byte c 5;
      stp_binout c a
  | Choice(a) ->
      output_byte c 6;
      stp_binout c a
  | True -> output_byte c 7
  | Neg -> output_byte c 8
  | Or -> output_byte c 9
  | And -> output_byte c 10
  | Iff -> output_byte c 11
  | Exists(a) ->
      output_byte c 12;
      stp_binout c a
  | DB(i,a) ->
      output_byte c 13;
      varint_binout c i;
      stp_binout c a
  | Lam(a,m) ->
      output_byte c 14;
      stp_binout c a;
      trm_binout c m
  | Ap(m,n) ->
      output_byte c 15;
      trm_binout c m;
      trm_binout c n

let my_input_byte c =
  let i = input_byte c in
(*  Printf.printf "[%d]" i; *) (* uncomment for debugging *)
  i

let rec varint_binin c =
  let i = my_input_byte c in
  if i <= 252 then
    i
  else if i = 253 then
    begin
      let j0 = my_input_byte c in
      let j1 = my_input_byte c in
      (j1 lsl 8) lor j0
    end
  else if i = 254 then
    begin
      let j0 = my_input_byte c in
      let j1 = my_input_byte c in
      let j2 = my_input_byte c in
      let j3 = my_input_byte c in
      (j3 lsl 24) lor (j2 lsl 16) lor (j1 lsl 8) lor j0
    end
  else
    let j = varint_binin c in
    (- j)
      
let string_binin c =
  let n = my_input_byte c in
  let b = Buffer.create 100 in
  for i = 1 to n do
    Buffer.add_char b (Char.chr (my_input_byte c))
  done;
  Buffer.contents b
    
let rec stp_binin c =
  match my_input_byte c with
  | 1 -> Prop
  | 2 -> Base(string_binin c)
  | 3 ->
      let a1 = stp_binin c in
      let a2 = stp_binin c in
      Ar(a1,a2)
  | _ -> raise (Failure "bad stp from channel")

let rec trm_binin c =
  match my_input_byte c with
  | 1 ->
      let x = string_binin c in
      let a = stp_binin c in
      Name(x,a)
  | 2 -> False
  | 3 -> Imp
  | 4 ->
      let a = stp_binin c in
      Forall(a)
  | 5 ->
      let a = stp_binin c in
      Eq(a)
  | 6 ->
      let a = stp_binin c in
      Choice(a)
  | 7 -> True
  | 8 -> Neg
  | 9 -> Or
  | 10 -> And
  | 11 -> Iff
  | 12 ->
      let a = stp_binin c in
      Exists(a)
  | 13 ->
      let i = varint_binin c in
      let a = stp_binin c in
      DB(i,a)
  | 14 ->
      let a = stp_binin c in
      let m = trm_binin c in
      Lam(a,m)
  | 15 ->
      let m = trm_binin c in
      let n = trm_binin c in
      Ap(m,n)
  | _ -> raise (Failure "bad trm from channel")

let lpar p = if p then "(" else ""

let rpar p = if p then ")" else ""

let rec stp_str_rec a p =
  match a with
    Base(x) -> x
  | Prop -> "prop"
  | Ar(b,c) -> (lpar p) ^ (stp_str_rec b true) ^ ">" ^ (stp_str_rec c false) ^ (rpar p)

let stp_str a = stp_str_rec a false

let rec trm_str_rec m p =
  match m with
    Name(x,_) -> x
  | False -> "False"
  | Imp -> "imp"
  | Forall(a) -> "Pi:" ^ (stp_str a)
  | Eq(a) -> "eq:" ^ (stp_str a)
  | Choice(a) -> "Sepsilon:" ^ (stp_str a)
  | DB(i,_) -> "^" ^ (string_of_int i)
  | Lam(a,m) -> (lpar p) ^ "\\_:" ^ (stp_str a) ^ "." ^ (trm_str_rec m false) ^ (rpar p)
  | Ap(m1,m2) ->
      begin
	match m1 with
	| Lam _ -> (lpar p) ^ (trm_str_rec m1 true) ^ " " ^ (trm_str_rec m2 true) ^ (rpar p)
	| _ -> (lpar p) ^ (trm_str_rec m1 false) ^ " " ^ (trm_str_rec m2 true) ^ (rpar p)
      end
(*** If logic has not been normalized away ***)
  | True -> "True"
  | Neg -> "not"
  | Or -> "or"
  | And -> "and"
  | Iff -> "iff"
  | Exists(_) -> "Sigma"

let trm_str m = trm_str_rec m false

let trm_ocaml_match m =
  let whenl = ref [] in
  let rec stp_ocaml_match_r a =
    match a with
    | Base("$i") ->
	let v = Printf.sprintf "v%d" (List.length !whenl) in
	whenl := Printf.sprintf "basetp_i_p %s" v :: !whenl;
	Printf.sprintf "Base(%s)" v
    | Base(x) ->
	let v = Printf.sprintf "v%d" (List.length !whenl) in
	whenl := Printf.sprintf "basetp_%s_p %s" x v :: !whenl;
	Printf.sprintf "Base(%s)" v
    | Prop -> "Prop"
    | Ar(a1,a2) -> Printf.sprintf "Ar(%s,%s)" (stp_ocaml_match_r a1) (stp_ocaml_match_r a2)
  in
  let rec trm_ocaml_match_r m =
    match m with
    | Name(x,a) ->
	let v = Printf.sprintf "v%d" (List.length !whenl) in
	if x.[0] = '_' then
	  whenl := Printf.sprintf "eigen%s_p %s" x v :: !whenl
	else
	  whenl := Printf.sprintf "%s_p %s" x v :: !whenl;
	Printf.sprintf "Name(%s,%s)" v (stp_ocaml_match_r a)
    | False -> "False"
    | Imp -> "Imp"
    | Forall(a) -> Printf.sprintf "Forall(%s)" (stp_ocaml_match_r a)
    | Eq(a) -> Printf.sprintf "Eq(%s)" (stp_ocaml_match_r a)
    | Choice(a) -> Printf.sprintf "Choice(%s)" (stp_ocaml_match_r a)
    | True -> "True"
    | Neg -> "Neg"
    | Or -> "Or"
    | And -> "And"
    | Iff -> "Iff"
    | Exists(a) -> Printf.sprintf "Exists(%s)" (stp_ocaml_match_r a)
    | DB(i,a) -> Printf.sprintf "DB(%d,%s)" i (stp_ocaml_match_r a)
    | Lam(a,m) -> Printf.sprintf "Lam(%s,%s)" (stp_ocaml_match_r a) (trm_ocaml_match_r m)
    | Ap(m,n) -> Printf.sprintf "Ap(%s,%s)" (trm_ocaml_match_r m) (trm_ocaml_match_r n)
  in
  let c = trm_ocaml_match_r m in
  match !whenl with
  | [] -> c
  | w::wr ->
      let rec andwhens wl s =
	match wl with
	| [] -> s
	| w::wr ->
	    andwhens wr (Printf.sprintf "%s && %s" s w)
      in
      andwhens wr (Printf.sprintf "%s when %s" c w)

